// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	AbsorptionDataServer.h --- class to retrieve absortion coefficients in CCF
//
//	DESCRIPTION:
//		A class AbsorptionDataServer is declared instance of which are to be
//		used to retrive absorption coefficient data from the CCF.
//		The query is keyword based.
//
//	CHANGE HISTORY:
//		UL, 1999-09-08: 2.0.0: - first version from CalAtomsXmm
//
//	AUTHOR:
//		U. Lammers, Fri Sep 10 11:16:48 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Calibration Access and Data Handbook, C. Erd et. al
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/AbsorptionDataServer.h,v 1.8 2003/10/03 14:15:40 sasbuild Exp $
//

#ifndef ABSORPTION_DATA_SERVER_H
#define ABSORPTION_DATA_SERVER_H

#include <vector>
#include <string>
#include <map>

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "slatec++.h"

//
//	---( AbsCoefCurves/AtomicProperties/AbsorptionDataServer )-----------------
//
typedef Slatec::CubicHermiteInterpolant<double>	AbsCoefCurve;

class AbsCoefCurves : public std::vector<AbsCoefCurve *> {
public:
	AbsCoefCurves() {};
	AbsCoefCurves(const std::string &name) : _name(name) {}

	std::string name() const { return _name; }

	unsigned coefficients() const { return this->size(); }

	const AbsCoefCurve &coefficient(unsigned n = 0) const;

	const AbsCoefCurve &f1() const { return coefficient(0); }
	const AbsCoefCurve &f2() const { return coefficient(1); }

	~AbsCoefCurves() {};

private:
	std::string	_name;
};


class AtomicProperties {
public:
	AtomicProperties() {};
	AtomicProperties(const std::string &name, unsigned number, double weight,
					 double density)
		: _name(name), _number(number), _weight(weight), _density(density) {}

	std::string name() const { return _name; }
	unsigned number() const { return _number; }
	double weight() const { return _weight; }
	double density() const { return _density; }

	~AtomicProperties() {}

private:
	std::string	_name;
	unsigned	_number;
	double		_weight, _density;
};


struct AbsorptionData {
	AbsorptionData() {}
	AbsorptionData(const AbsCoefCurves &absCoefCurves,
				   const AtomicProperties &atomicProperties)
		: henke(absCoefCurves), atomic(atomicProperties) {}

	AbsCoefCurves		henke;
	AtomicProperties	atomic;
};


class AbsorptionDataServer : public CalAtom,
		   public virtual CalChild<CalAtomCollection<AbsorptionDataServer> >,
		   public std::map<std::string, AbsorptionData, less<std::string> > {
public:
	AbsorptionDataServer(CalAtomCollection<AbsorptionDataServer> *parent = 0)
		: CalChild<CalAtomCollection<AbsorptionDataServer> >(parent) {}
		//
		//	PURPOSE:
		//		Instantiate an AbsorptionDataServer object
		//


	static unsigned id() { return Ccf::AbsCoefs; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	virtual std::vector<std::string> materials() const;
		//
		//	PURPOSE:
		//		Return the list of materials for which absoption curves
		//		are known to the object
		//


	virtual bool hasMaterial(const std::string &materialSymbol) const;
		//
		//	PURPOSE:
		//		Check whether an absorption curve for a specified material
		//		exists
		//
		//	ARGUMENTS:
		//		materialSymbol	: Id of the material as in the Periodic Table
		//						  of elements (e.g. "Si", "O", "F", etc.)	
		//


	virtual const AbsCoefCurves &henke(const std::string &materialSymbol)const;
		//
		//	PURPOSE:
		//		Retrieve the absorption coefficient curves for a specified
		//		material; Please note: attempting to retrieve a non-existing
		//		data item will result in an error condition.
		//
		//	ARGUMENTS:
		//		materialSymbol	: Id of the material as in the Periodic Table
		//						  of elements (e.g. "Si", "O", "F", etc.)	
		//


	virtual const AtomicProperties &atomicData(const std::string &
																materialSymbol)
	const;
		//
		//	PURPOSE:
		//		Retrieve atomic property data for a specified
		//		material; Please note: attempting to retrieve a non-existing
		//		data item will result in an error condition.
		//
		//	ARGUMENTS:
		//		materialSymbol	: Id of the material as in the Periodic Table
		//						  of elements (e.g. "Si", "O", "F", etc.)	
		//


	virtual ~AbsorptionDataServer();
		//
		//	PURPOSE:
		//		Destrory AbsorptionDataServer object constructed with above
		//		constructor(s).
		//

private:
	typedef std::map<std::string, AbsorptionData, std::less<std::string> >
			MapType;
};
	
#endif
