// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmFluxConverter.cc --- implementation of OmFluxConverter class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2000-01-27: 1.0.0: - initial version
//		UL, 2000-03-03: 1.1.0: - OmFluxConverter::fluxCorrect(): brought
//								 implementation in line with latest
//								 description in CAL HB
//		UL, 2002-01-28: 1.2.0: - fluxCorrect(): estimate errors of converted
//								 count rate (does not take into account
//								 coincidence loss correction)
//		UL, 2002-04-24: 1.3.0: - fixed bug in calculation of converted count
//								 rate
//
//	AUTHOR:
//		U. Lammers (UL), Thu Jan 27 11:39:07 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/OmFluxConverter.cc,v 1.18 2009/07/06 14:25:19 aibarra Exp $
//

#include <cfloat>
#include <cmath>
#include <climits>

#include "CalOalUtility.h"
#include "Evaluator.h"
#include "OmFluxConverter.h"

using namespace std;

//
//	---( OmFluxConverter> ) ---------------------------------------
//
OmFluxConverter::OmFluxConverter(const CcfConstituentTable &photToNatTab,
					CalAtomCollection<OmFluxConverter> *parent)
	: CalChild<CalAtomCollection<OmFluxConverter> >(parent)
{
	init(photToNatTab);
} // OmFluxConverter::OmFluxConverter()


OmFluxConverter::OmFluxConverter(const CcfConstituentTable &photToNatTab,
					const CcfConstituentTable &coincidenceTab,
					CalAtomCollection<OmFluxConverter> *parent)
	: CalChild<CalAtomCollection<OmFluxConverter> >(parent)
{
	init(photToNatTab, &coincidenceTab);
} // OmFluxConverter::OmFluxConverter()

OmFluxConverter::OmFluxConverter(const CcfConstituentTable &photToNatTab,
					const CcfConstituentTable &coincidenceTab,
					const std::vector<double> &degradCoeffs,
                                        const double refEpoch,
                                        const double mjd_peak,
                                        const double cst,
                                        const std::vector<OmFluxConverter::EvalTuple> &region,
					std::vector<Slatec::LinearInterpolant<double> *> &posDepinterp,
                                        bool filter_data, bool posdep_data, 
                                        unsigned algoid,
					CalAtomCollection<OmFluxConverter> *parent)
	: CalChild<CalAtomCollection<OmFluxConverter> >(parent),
      _degradCoeffs(degradCoeffs), _refEpoch(refEpoch), _mjd_peak(mjd_peak), _cst(cst), _region(region), _posDepinterp(posDepinterp), _filter_data(filter_data), _posdep_data(posdep_data), _algoid(algoid)
{
	//init(photToNatTab, &coincidenceTab, &degradCoeffs);
	init(photToNatTab, &coincidenceTab);
} // OmFluxConverter::OmFluxConverter()



void OmFluxConverter::init(const CcfConstituentTable &photToNatTab,
						   const CcfConstituentTable *coincidenceTab)
						 //  const std::vector<double> *degradCoeffs)
{
   // _degradCoeffs=degradCoeffs;
	const Table *table = photToNatTab.block();

	//
	//	read attributes
	//
	_phdFrac = table->attribute("PHDFRAC")->asReal(),
	_thresh	= table->attribute("THRESH")->asReal();

	//
	//	read table
	//
	ColumnData	*cdparam = table->column("PARAM00")->data(),
				*cdparame = table->column("DPARAM00")->data();
	real32 *param = cdparam->real32Data(),
		   *parame = cdparame->real32Data();

	if (cdparam->elements() < 8)
		errstr<<"Need a least 8 elements in column "
			  <<table->column("PARAM00")->qualifiedName()
			  <<Sas::error("TooLittleData");
	if (cdparam->elements() != cdparame->elements())
		errstr<<"Number of elements in columns "
			  <<table->column("PARAM00")->qualifiedName()
			  <<" and "
			  <<table->column("DPARAM00")->qualifiedName()
			  <<" do not match"
			  <<Sas::error("VectorLengthMismatch");

	OmFluxConverter::FrameParameters &fp = _frameParameters;
	fp.nHorizontalPixels.value = param[cdparam->position(0, 0)];
	fp.nHorizontalPixels.error = parame[cdparame->position(0, 0)];
	fp.nVerticalPixels.value = param[cdparam->position(0, 1)];
	fp.nVerticalPixels.error = parame[cdparame->position(0, 1)];
	fp.verticalTransferTime.value = param[cdparam->position(0, 2)];
	fp.verticalTransferTime.error = parame[cdparame->position(0, 2)];
	fp.horizontalTransferTime.value = param[cdparam->position(0, 3)];
	fp.horizontalTransferTime.error = parame[cdparame->position(0, 3)];
	fp.nVerticalPixelsImage.value = param[cdparam->position(0, 4)];
	fp.nVerticalPixelsImage.error = parame[cdparame->position(0, 4)];
	fp.verticalHorizontalSwitchTime.value = param[cdparam->position(0, 5)];
	fp.verticalHorizontalSwitchTime.error = parame[cdparame->position(0, 5)];
	fp.horizontalVerticalSwitchTime.value = param[cdparam->position(0, 6)];
	fp.horizontalVerticalSwitchTime.error = parame[cdparame->position(0, 6)];
	fp.redundantCcdOffset.value = param[cdparam->position(0, 7)];
	fp.redundantCcdOffset.error = parame[cdparame->position(0, 7)];

	//	tidy up
	delete cdparam;
	delete cdparame;

	//	read COINCIDENCE table if present
	if (coincidenceTab)
		CalOalUtility::fillVector(coincidenceTab->block(), "PLINFUNC", 0,
								  _linearityCoeff);

	//	Set degradation coefficients for this filter if available
	//if (degradationTab)
	//	CalOalUtility::fillVector(degradationTab->block(),
      //                            uppercase(Xmm::toString(
        //                             calServer.state()->filter()->get())), 0,
          //                        _degradCoeffs);

} // OmFluxConverter::init()


void OmFluxConverter::dump(ostream &os)
{
	os<<"OmFluxConverter{\n";
	os<<"}";
} // OmFluxConverter::dump()


OmFluxConverter::FrameParameters OmFluxConverter::frameParameters() const
{
	return _frameParameters;
} // OmFluxConverter::frameParameters()


double OmFluxConverter::phdFrac() const
{
	return _phdFrac;
} // OmFluxConverter::phdFrac()


double OmFluxConverter::thresh() const
{
	return _thresh;
} // OmFluxConverter::thresh()

vector<double> OmFluxConverter::degradationCoeffs() const
{
   // Issue a warning and set to zero if not available
    if (_degradCoeffs.size() < 1) {
        errstr<<"DEGRADATION coefficients not available from PHOTTONAT CCF"
              <<Sas::warning("MissingDegradationTable");
        std::vector<double> dc;
        for (unsigned n=0; n<3; ++n)
            dc.push_back(0.0);

        return dc;
    }
    else
	    return _degradCoeffs;

} // OmFluxConverter::degradationCoeffs()

/*
* Method: calcDegradationFactor
* Purpose: calculates the fraction of flux lost due to degradation 
*          at a given MJD for a given filter. This calculation used to
*          be performed in mssllib (scr-7552)
*/
const double OmFluxConverter::calcDegradationFactor(double MJD) const
{
  // Initial algorithm
  double dfac=1.0;

  // Only apply correction for filters where data exists in CCF
  if (_filter_data) {

    if (_algoid==0) 
    {
     /* Polynomial algo */
      dfac=0.0;
      for (unsigned i=0; i<_degradCoeffs.size(); i++)
         dfac = dfac + _degradCoeffs[i] * pow(MJD,i); 
    }
    else if (_algoid==1)
    {
     /* This is the exponential degradation function where:
        D(t) = 1 / (f x exp[-(t-t0)/td] + (1-f))
	degradCoeffs[0]=f, [1]=td (in days), [2]=t0 (as an MJD)
     */
         double f=_degradCoeffs[0], td=_degradCoeffs[1], t0=_degradCoeffs[2];
         dfac = 1.0 / (f * exp(-(MJD-t0)/td) + (1.0-f));
         //cout<<" f MJD td t0 dfac "<<f<<" "<<MJD<<" "<<td<<" "<<t0<<" "<<dfac;
    }

    // This is the algo which was never implemented as a CCF
    //else { // Use new algorithm - SCR-7555 
     // NEED TO ENSURE THAT THIS IS FORWARDS COMPATIBLE !!

     // Calculate the offset time from the epoch when degradation started
     //double t = MJD - _refEpoch;

     // The factor is calculated using a polynomial up to a certain date
     // and is thereafter set to a constant
     //if (MJD <= _mjd_peak)
       // for (unsigned i=0; i<_degradCoeffs.size(); i++)
         //  dfac = dfac + _degradCoeffs[i] * pow(t,i);
    // else 
      //  dfac =_cst; 
    //}
  }
  //cout<<_algoid<<" "<<dfac<<" "<<_refEpoch<<" "<<_cst<<" "<<_mjd_peak<<" "<<MJD<<" "<<endl;

  return dfac;

} // OmFluxConverter::calcDegradationFactor

/*
* Method: posDepDegradationFactor
* Purpose: calculates the fraction of flux lost due to degradation 
*         at a given MJD for a given filter at a given position on the detector. 
*         This value should be multiplied by the position independent factor
*         given by calcDegradationFactor
*/
const double OmFluxConverter::posDepDegradationFactor(double mjd,
          double detx, double dety) const
{
  // Initial algorithm
  double dfac=1.0;

  // Only apply correction for filters where pos depnt factors exist 
  if (_posdep_data && _filter_data) {

     // Check whether the supplied position lies within a region
     for (unsigned r=0; r<_region.size(); r++) {
         *_region[r].chipX = detx;
         *_region[r].chipY = dety;
         if (_region[r].evaluator->evalLogic())  {
            // Calculate the interpolated correction factor for the
            // given date. Don't apply corection for dates eariler
            // than the initial calibration point but extrapolate for
            // dates after the final point
            if (mjd < _region[r].minMjd)
              dfac = 1.0;
            else
              dfac = (*_posDepinterp[r])(mjd);
         }
     }
  }

  return dfac;
} // OmFluxConverter::posDepDegradationFactor


ValueAsymmetricError OmFluxConverter::fluxCorrect(ValueAsymmetricError rawRate,
									  double frameTime,
									  double deadFrac,
									  double eventLoss,
									  bool extendedSource,
									  bool empiricalLinearityCorrection) const
{
	double rate = rawRate.value;
	ValueAsymmetricError res = rawRate;

	//
	//	coincidence loss correction (only for non-extended sources)
	//
	if (!extendedSource) {
		if (frameTime == 0.)
			errstr<<"Frametime is 0 - cannot correct raw OM count rate"
				  <<Sas::error("IllegalFrameTime");

		if (rate < 1./frameTime) {
			double f = 1./(1.-rate*frameTime);
			rate = -log(1.-rate*frameTime)/frameTime;
			res.error.errorPlus = f*rawRate.error.errorPlus;
			res.error.errorMinus = f*rawRate.error.errorMinus;
		} else {
			res.value = DBL_MAX;
			res.error.errorPlus = res.error.errorMinus = DBL_MAX;
			errstr<<"The count rate is greater than the frame rate - this is "
					"not possible - cannot correct raw OM count rate"
			  <<Sas::warning("IllegalCountRate");
			return res;
		}

		//	apply empirical correction if requested
		unsigned N = _linearityCoeff.size();
		if (empiricalLinearityCorrection && N) {
			unsigned j=N-1;
			double v = _linearityCoeff[j],
				   x = rawRate.value * frameTime;
			while (j>0)
				v = v*x + _linearityCoeff[--j];
			rate *= v;
		}
	}

	//
	//	correct for deadtime
	//
	if (deadFrac == 1.)
		errstr<<"Deadtime is 1 - cannot correct raw OM count rate"
			  <<Sas::error("IllegalDeadTime");
	rate /= 1.-deadFrac;

	//
	//	correction for average detector aging
	//
	/*
	 * AI gcc4 porting
	 * suggest explicit braces to avoid ambiguous 'else'
	 * => ()
	 */
	if (eventLoss != 0.)
	{
		if (eventLoss == 1.)
			errstr<<"Fraction of events below threshold is 1 - cannot "
					"correct raw OM count rate"
				  <<Sas::error("IllegalEventLossRate");
		else
			rate /= (1.-eventLoss) * _phdFrac;
	}
	res.value = rate;

	return res;
} // OmFluxConverter::fluxCorrect()

OmFluxConverter::~OmFluxConverter()
{
} // OmFluxConverter::~OmFluxConverter()


//
//	---( CalAtomCollection<OmFluxConverter> ) ----------------------
//
/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<OmFluxConverter>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	OmFluxConverter *null = 0;
	_collection = new vector<OmFluxConverter *>(_nEntries, null);
	OmFluxConverter *&ff = (*_collection)[0];

	//
	//	access data tables
	//
	CifEntry *e = ccf->entry(instrument, Ccf::PhotToNat);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet, "PHOTTONAT");
	if (!dataSet->hasBlock("COINCIDENCE")) {
		errstr<<"Data set `"<<dataSet->qualifiedName()<<"' does not possess "
				"a table COINCIDENCE - the empirical rate linearity correction"
				" will have no effect"
			  <<Sas::warning("MissingCoincidenceTable");
		ff = new OmFluxConverter(ccftab, this);
	} else {

	//	Set degradation coefficients for this filter if available
        std::vector<double>     degradCoeffs;

        // Use the exponential coefficients if they exist, otherwise use
        // the polynomial coeffs
        if (dataSet->hasBlock("EXPDEGRAD") || dataSet->hasBlock("DEGRADATION")) 
        {
           string degblock;
           unsigned maxalg=0;
           if (dataSet->hasBlock("EXPDEGRAD")) {
              degblock="EXPDEGRAD";
              maxalg=1;
           } else if (dataSet->hasBlock("DEGRADATION")) {
              degblock="DEGRADATION";
              maxalg=0;
			  // Issue a warning about old style CCF
			  errstr<<dataSet->qualifiedName()<<" uses an outdated "
				"OM filter flux degradation correction method. It is recommended to"
				"update your CCF to the latest version"<<Sas::warning("OutOfDateCCF");
           }
           CcfConstituentTable degradationTab(dataSet, degblock, maxalg);

          // Read the degradation coefficients into an array
           std::vector<double>     degradCoeffs;
           // Open degradation extension 
           unsigned algoid = degradationTab.algoid();

        
           double refEpoch=0, mjd_peak=0, cst=0;
           bool filter_data=true;

          // If degradation coefficients exist for this filter read them
           if (degradationTab.block()->hasColumn(
                               uppercase(Xmm::toString(filter)))) {

              CalOalUtility::fillVector(degradationTab.block(),
                                        uppercase(Xmm::toString(filter)), 0,
                                        degradCoeffs);
              /* Get keywords if required - no longer needed
              if (algoid >=1) {
                 refEpoch=degradationTab.block()->attribute("REFEPOCH")
                                                               ->asReal();
                 mjd_peak=degradationTab.block()->attribute(
                       uppercase(Xmm::toString(filter))+"_TMX")->asReal();
                 cst=degradationTab.block()->attribute(
                       uppercase(Xmm::toString(filter))+"_CST")->asReal();
              } */
 
           } else {
              // Set filter data flag to false
              filter_data=false;
           }

          // Read in position dependent degradation factors if they exist
           bool posdep_data = false;
           std::vector<OmFluxConverter::EvalTuple> region;
           vector<Slatec::LinearInterpolant<double> *> posDepinterp;

          // Read them in if the FILTER data is present - will be same filters
          // as the time-dependent correction
           if (dataSet->hasBlock("BORECORRFACTORS") && filter_data) {
              posdep_data = true;

              // Open the pos dept correction extension
              CcfConstituentTable bcftab(dataSet, "BORECORRFACTORS", 0);
              const Table *table = bcftab.block();

              std::vector<double> pdcmjd;
              std::vector<double> posdepfactors;

              // Loop over each record
              ColumnData *cdr = table->column("REGION")->data();
              for (unsigned n=0; n<table->rows(); ++n) {

                // Read the dates and corr factors for this filter/region
                CalOalUtility::fillVector(table, "MJD", n,pdcmjd);
                string factorCol = uppercase(Xmm::toString(filter))+
                                                    "_BORECORRFACTOR";
                CalOalUtility::fillVector(table, factorCol, 
                                                   n, posdepfactors);
                // Create a linear interpolator for the pos dep factors
                // as a function of MJD
                posDepinterp.push_back(new Slatec::LinearInterpolant
                             <double>(pdcmjd, posdepfactors));

                // Make a region evaluator for this region
                string regstr(stringColumnRow(cdr, n));

                if (!Evaluator::isValidExpression(regstr))
                  errstr<<"The region specification `"<<regstr<<"' in `"
                  <<cdr->parent()->qualifiedName()<<"' is syntactically "
                    "incorrect"<<Sas::error("CcfDataIntegrityError");

                OmFluxConverter::EvalTuple et;
                et.evaluator = new Evaluator(regstr);
                NumericSymbols syms = et.evaluator->numericIdentifiers();
                et.chipX = syms.valuePtr[0];
                et.chipY = syms.valuePtr[1];
                et.minMjd = *std::min_element(pdcmjd.begin(),
                                       pdcmjd.end()); // Set min MJD 

                // Add this region into the region vector
                region.push_back(et); 

              }
              delete cdr;
           }              

           // Initialise the flux converter
           ff = new OmFluxConverter(ccftab,
                                 CcfConstituentTable(dataSet, "COINCIDENCE"),
                                 degradCoeffs, refEpoch, mjd_peak, cst,
                                 region, posDepinterp, filter_data, 
                                 posdep_data, algoid, this);
        } else
		   ff = new OmFluxConverter(ccftab,
								 CcfConstituentTable(dataSet, "COINCIDENCE"),
								 this);
    }

	dataSetServer->close(dataSet);
} // CalAtomCollection<OmFluxConverter>::CalAtomCollection()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
OmFluxConverter *CalAtomCollection<OmFluxConverter>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<OmFluxConverter>::select
